Option Explicit On 
Option Strict On


'#############################################################################################
' John Girard, Advanced Systems & Controls
'
' HandlerCollection: a simple collection extension that contains a collection of exception 
' handler-compatible objects. this collection also includes a few routines to broadcast errors
' or warnings to all members of this collection. this collection also implements the iHandler 
' interface itself.
'
' Interfaces Required:  iHandler
' Classes Required:     State
' Modules Required:     none 
' Other Requirements:   none
'
' Certification:        In Progress (05/21/04)
'                       Module Active in 0 Projects
'
' Revision History:
' --------------------------------------------------------------------------------------------
'
' (08/13/03)    Reviewed code and comments for errors and ommissions
'
' (08/14/03)    Tested the file for 40 minutes, set the file to read-only and started the 
'               code certification process.
'
' (10/20/03)    Deleted a few unnecessary comment headers.
'
' (02/11/04)    Reformatted comment bars and comments.
'
' (03/12/04)    MILESTONE: class shipped with project ASC2288.
'
' (04/24/04)    Altered the Add routine: if an exact object exists in the Handler collection,
'               then it cannot be re-added (similar to how Concatenate works), to prevent
'               duplicate handlers. If the item exists, the Add routine returns that item's
'               current number instead of creating it.
'
' (04/28/04)    Tested the file for an additional 5 minutes to restart certification.
'
' (05/06/04)    Began a major rewrite to adopt an n-state based system rather than the simple
'               two-state error-based system. This entailed writing brand new interfaces, 
'               altering all handler-based classes, etc.
'
' (05/07/04)    Finished the n-state rewrite. Double-checked and tested the code for about
'               15 minutes and restarted certification.
'
' (05/12/04)    MILESTONE: class shipped with project ASC DrawMove (internal).
'
' (05/20/04)    Continuing with the exception handler rewrite, implemented the new persistent
'               State class for this class. Reviewed and tested the class for 5 minutes and
'               restarted the certification process.
'
' (05/21/04)    Added the new Force sub calls (instead of the old simple Change calls for 
'               the states) to force transitions for each individual device. The most changes 
'               occurred in the State class.
'
'#############################################################################################

Namespace naExc

    Public Class HandlerCollection
        Inherits CollectionBase
        Implements naExc.iHandler

        'the state declaration; this is typically created by constructor
        Private WithEvents omState As New naExc.State


        '#####################################################################################
        'default property: return the item specified or assign a new one. since the exception 
        'handlers are relatively stand-alone, this won't be accessed too often except in
        'special cases
        '#####################################################################################
        Default Public Property Item(ByVal iIndex As Integer) As naExc.iHandler

            Get
                Return CType(MyBase.List(iIndex), naExc.iHandler)
            End Get

            Set(ByVal Value As naExc.iHandler)
                MyBase.List(iIndex) = Value
            End Set

        End Property


        '#####################################################################################
        'add a new item to this collection, and return a simple integer key that can be used
        'to reference the item. only add an item if it does not already exist. in addition,
        'when a new handler is added, we must assign the current state to the new handler
        '#####################################################################################
        Public Function Add(ByVal oValue As naExc.iHandler) As Integer
            Dim I As Integer

            'first, set the new handler's state equal to the current state. force an edge
            'transition since the current handler's state must be different.
            oValue.State.ForceEdgeTransition(omState)

            'if our item exists, find its number and return that as its reference
            If MyBase.List.Contains(oValue) Then

                'find the reference and immediately return it
                For I = 0 To Me.Count - 1
                    If Me.Item(I) Is oValue Then Return I
                Next
            Else
                'otherwise, create a new item and return its number
                Return MyBase.List.Add(oValue)
            End If

        End Function


        '#####################################################################################
        'add all members of the passed exception handler to this exception handler. if the 
        'passed handler collection contains a reference to a handler that exists in this 
        'collection, it is not duplicated
        '#####################################################################################
        Public Sub Concatenate(ByRef oHnd As HandlerCollection)
            Dim oCur As naExc.iHandler

            'loop through all exception handlers in the passed collection. only add items 
            'whose references do not exist in the current collection. new handlers are
            'automatically assigned the correct state by the Add function
            For Each oCur In oHnd
                If Not MyBase.List.Contains(oCur) Then MyBase.List.Add(oCur)
            Next

        End Sub


        '#####################################################################################
        'remove the specified item from collection
        '#####################################################################################
        Public Sub Remove(ByVal oValue As naExc.iHandler)
            MyBase.List.Remove(oValue)
        End Sub


        '#####################################################################################
        'does this collection contain the passed reference?
        '#####################################################################################
        Public Function Contains(ByVal oValue As naExc.iHandler) As Boolean
            Return MyBase.List.Contains(oValue)
        End Function


        '#####################################################################################
        'assign the default state to this handler collection. setting a new state
        'automatically passes down the state to all handlers included. note that changing
        'the state's properties will NOT update the state; a whole new state must be created
        '#####################################################################################
        Public ReadOnly Property State() As naExc.State Implements iHandler.State

            Get
                Return omState
            End Get

        End Property


        '#####################################################################################
        'outer code has reported a simple message. pass it on down the line to the other 
        'handlers
        '#####################################################################################
        Public Sub debugText(ByVal sText As String) Implements naExc.iHandler.debugText
            Dim oHandler As naExc.iHandler

            'pass the debug message down to all items
            For Each oHandler In MyBase.List
                oHandler.debugText(sText)
            Next

        End Sub


        '#####################################################################################
        'pass an edge transition down to all other state handlers
        '#####################################################################################
        Private Sub omStateEdgeTransition(ByVal oSender As Object, ByVal eArgs As System.EventArgs) Handles omState.EdgeTransition
            Dim oHandler As naExc.iHandler

            For Each oHandler In MyBase.List
                oHandler.State.ForceEdgeTransition(Me.State)
            Next
        End Sub


        '#####################################################################################
        'same general concept for level transitions too
        '#####################################################################################
        Private Sub omStateLevelTransition(ByVal oSender As Object, ByVal eArgs As System.EventArgs) Handles omState.LevelTransition
            Dim oHandler As naExc.iHandler

            For Each oHandler In MyBase.List
                oHandler.State.ForceLevelTransition(Me.State)
            Next
        End Sub

    End Class

End Namespace