Option Explicit On 
Option Strict On


'#############################################################################################
' John Girard, Advanced Systems & Controls
'
' IntensityMap: a container for an array of color integers and intensities.
'
' Revision History:
' --------------------------------------------------------------------------------------------
'
' (02/08/05)    Initial creation of the class.
'
'#############################################################################################

Public Class IntensityMap
    Inherits ColorImage

    Public Color(,) As Integer
    Public Intensity(,) As Double


    '#########################################################################################
    'create a new image. all color arrays have the same size and depth
    '#########################################################################################
    Sub New(ByVal iWidth As Integer, ByVal iHeight As Integer)
        MyBase.New(iWidth, iHeight)
    End Sub


    '#########################################################################################
    'resize the image. data is never preserved during a resize.
    '#########################################################################################
    Protected Overrides Sub Resize()

        ReDim Color(imWidth - 1, imHeight - 1)
        ReDim Intensity(imWidth - 1, imHeight - 1)

    End Sub



    '#########################################################################################
    'given a collection of colors to isolate and the red-blue chrominance maps of an image, 
    'convert the chrominance arrays into a complement: first array will contain a map with 
    'each element corresponding to the strongest color, while the second array will contain 
    'intensity of that particular color at that pixel (where 1 is a perfect match, 0 is no 
    'match)
    '#########################################################################################
    Public Sub FromYUV(ByVal fYUV As YUVImage, ByRef cColors As Collection)
        Dim I As Integer
        Dim J As Integer
        Dim fColor As FuzzyColor

        Dim iColor As Integer
        Dim dIntensity As Double
        Dim dCur As Double

        'at what (low) intensity do we declare that there is no match?
        Const INTENSITY_THRESHOLD As Double = 0.2

        'create the new intensity and color maps
        ReDim Color(fYUV.Width - 1, fYUV.Height - 1)
        ReDim Intensity(fYUV.Width - 1, fYUV.Height - 1)

        'loop through every single pixel in the map
        For I = 0 To fYUV.Width - 1
            For J = 0 To fYUV.Height - 1

                'reset current color and intensity
                iColor = 0
                dIntensity = 0

                'check the current pixel against all color probabilities
                For Each fColor In cColors
                    'get current probability. if it's better than current, then it replaces
                    'the current value.
                    dCur = fColor.Probability(fYUV.Y(I, J), fYUV.U(I, J), fYUV.V(I, J))
                    If dCur > dIntensity Then
                        iColor = fColor.ID
                        dIntensity = dCur
                    End If
                Next

                'below a certain confidence level, we just declare this value to
                'be a non-match.
                If dIntensity < INTENSITY_THRESHOLD Then
                    'once all colors have been checked, write the correct values
                    'to the array.
                    Color(I, J) = 0
                    Intensity(I, J) = 0
                Else
                    'once all colors have been checked, write the correct values
                    'to the array.
                    Color(I, J) = iColor
                    Intensity(I, J) = dIntensity
                End If
            Next
        Next

    End Sub


    '#############################################################################################
    'construct a segmentation map of the passed-in, filtered chrominances. the input array
    'consists of integers relating to a single color match. values set to zero represent no
    'match, while other values represent a single color match. only like colors are regioned
    'together.
    '#############################################################################################
    Public Function Regionalize(ByVal iMaxColor As Integer, ByRef iColor(,) As Integer, ByVal dIntensity(,) As Double) As Collection
        Dim I As Integer
        Dim J As Integer
        Dim K As Integer
        Dim rCur As Region
        Dim oRegions As New Collection

        Dim iLabels() As Integer

        'create an array that can serve as a reference for the labeling process
        ReDim iLabels(iMaxColor - 1)

        '*****************************************************************************************
        'loop through every single pixel in the image
        For I = 0 To iColor.GetLength(0) - 1
            For J = 0 To iColor.GetLength(1) - 1

                'does the current pixel have a color which is not labeled?
                If iColor(I, J) > 0 And iColor(I, J) <= iMaxColor Then

                    'create a new label for this pixel
                    iLabels(iColor(I, J) - 1) += 1
                    rCur = New Region(iColor(I, J))

                    'now, flood fill the entire region which matches it
                    FloodFill(rCur, iColor, dIntensity, iColor(I, J), iColor(I, J) + iMaxColor * iLabels(iColor(I, J) - 1), I, J)

                    'normalize centroid data
                    rCur.X /= rCur.Strength
                    rCur.Y /= rCur.Strength
                    oRegions.Add(rCur)
                End If
            Next
        Next

        Return oRegions

    End Function


    '#############################################################################################
    'construct a segmentation map of the passed-in, filtered chrominances. the input array
    'consists of integers relating to a single color match. values set to zero represent no
    'match, while other values represent a single color match. only like colors are regioned
    'together.
    '#############################################################################################
    Protected Sub FloodFill(ByRef rCur As Region, ByRef iImage(,) As Integer, ByRef dIntensity(,) As Double, ByVal iColor As Integer, ByVal iLabel As Integer, ByVal X As Integer, ByVal Y As Integer)

        'make sure we're within bounds for this flood fill
        If X > -1 And X < iImage.GetLength(0) And Y > -1 And Y < iImage.GetLength(1) Then

            'if we're within bounds, then we can check color.
            If iImage(X, Y) = iColor Then
                iImage(X, Y) = iLabel

                'add data about the current point to region statistics
                rCur.X += X * dIntensity(X, Y)
                rCur.Y += Y * dIntensity(X, Y)
                rCur.Strength += dIntensity(X, Y)
                rCur.Count += 1

                'is this the maximum value?
                If rCur.hiStrength < rCur.Strength Then
                    rCur.hiX = X
                    rCur.hiY = Y
                    rCur.hiStrength = rCur.Strength
                End If

                'flood fill in all four directions
                FloodFill(rCur, iImage, dIntensity, iColor, iLabel, X + 1, Y)
                FloodFill(rCur, iImage, dIntensity, iColor, iLabel, X, Y + 1)
                FloodFill(rCur, iImage, dIntensity, iColor, iLabel, X - 1, Y)
                FloodFill(rCur, iImage, dIntensity, iColor, iLabel, X, Y - 1)
            End If
        End If

    End Sub

End Class
