using System;

//############################################################################################
// John Girard, Advanced Systems & Controls
//
// Tooth: takes as input a signal consisting of a repeating square waveform, and formats this 
// input into a set of single cycle lists exported to a collection of Tooth classes.
// 
// Interfaces Required: none
// Classes Required:    none
// Modules Required:    none 
// Other Requirements:  none
//
// Certification:       NOT In Progress (12/30/03)
//                      Module Active in 0 Projects
//
// Revision History:
// -------------------------------------------------------------------------------------------
// 
// ( INITIAL)	Verified and double-checked. Officially added to the object tree. Added the 
//				xmlTable routine, which takes in an XML Excel header template, and outputs an 
//				Excel-compatible spreadsheet in XML format.
//
// (02/19/04)   Converted the class from VB6 to C# .NET format. Removed many of the extra
//				frills from the class and reduced it to a much more streamlined format.
//
//############################################################################################

namespace nfABS
{
	//create a simple enumeration for choosing the type of calculations used for getting
	//min and max points
	public enum eCuspType
	{
		Average, Quadratic
	};

	//create a simple class for storing a point's value
	public struct Point
	{
		public double X;
		public double Y;
	};

	public class Teeth
	{
		//assume the peak-to-peak distance for the cycle is 1.0, that the minimum is assigned
		//a relative Y value of 0.0, and the maximum is assigned a relative Y value of 1.0.
		//TTL high and TTL low represent the distances at which the wave starts and finishes
		//its rise, and starts and finishes its fall.
		public double ttlHi = 0.9;
		public double ttlLo = 0.1;

		//Y mean is an absolute positional value (in whatever units data is provided) that
		//determines where zero cross is.
		public double yMean = 0;
		
		//what sort of calculation should be used to calculate the exact values of the
		//local extrema?
		public eCuspType CuspType = eCuspType.Average;

		//X cut is a relative data value used to help determine minimum and maximum points.
		//with a range of 0 to 1, it represents the percentage of data between two zero
		//cross points that will be ignored while finding the maximum. for a wave like a
		//square wave, since the maximum and minimum are maintained for a long period, this
		//value should be set relatively low (so only the small rise and fall time bands
		//will be ignored. for a wave like a sine wave, where minimum and maximum only truly
		//peak at one point, this value should be set relatively high (so only real max and
		//min data is factored in).
		public double xCut = 0.8;

		//Y cut is an absolute data value used to determine hysteresis band. basically, it
		//determines the amount of band around the zero cross point (yMean) that will be
		//used to find the exact zero cross point.
		public double yCut = 0.1;

		//####################################################################################
		// given the passed array, build a table of primitive, basic calculations about the
		// waveform contained in the array.
		//####################################################################################
		public unsafe nfABS.Point[] Process(double[] dData)
		{
			//what type of transition do we need? true represents a low-to-high transition
			//while false represents a high-to-low transition
			bool bTrans = false;
			int I, J = 0, K;

			//the three main arrays containing all landmark points for the signal
			Point[] dMid;
			Point[] dExt;
			Point[] dTTL;

			//is the code currently inside a hysteresis band cut?
			bool bInCut = false;
			
			//has a good set of points (I to J) been generated and can we now find the
			//midpoint?
			bool bCalc = false;

			int iLo;
			int iHi;

			double dLo;
			double dHi;
			double dDelta;

			//a useful set of pointers for blazing fast array operations
			double* pS;
			double* pJ;
			double dTemp;

			//variables used for building the linear regression model
			double[] dList;
			nfMath.PolyRegression oReg;
			nfMath.Matrix mParam;
			Point pCur;

			//a temporary linked list for data calculation
			nfMath.LinkedList oL;

			//the linked list containing all zero cross midpoints
			nfMath.LinkedList oMid = new nfMath.LinkedList();

			//set values for just below the hysteresis (horizontal) band and for just
			//above it
			dLo = yMean - 0.5 * yCut;
			dHi = yMean + 0.5 * yCut;

			//fix the locations of the source array so the garbage collector cannot move 
			//it while the code is being executed.		
			fixed (double* pSrc = dData)
			{
				//assign pointer to beginning of the list
				pS = pSrc;
				pJ = pSrc;

				//iterate through all points except the first one (since we want to
				//be able to access the point before the current one)
				for(I = 1; I < dData.Length; I++)
				{
					//increment through the list automatically; this will start the
					//list at item 1 (instead of item 0)
					pS++;

					//are we currently looking for a high-to-low transition?
					if(bTrans == false)
					{
						//was there a transition from high to inside? then set the "inside" 
						//flag to true and hold on. or, was there a transition from inside to
						//high? in that case, we've left the hysteresis band without actually 
						//getting a zero cross point, so reset the "inside" flag.
						if(*(pS - 1) >= dHi && *pS < dHi)
						{
							//use the point just outside the transition as the first point 
							//used for calculation. save both a pointer to the current item 
							//and a counter reflecting the current item
							pJ = pS - 1;
							J = I - 1;
							bInCut = true;
						}
						else if(*(pS - 1) < dHi && *pS >= dHi && bInCut == true)
						{
							bInCut = false;
						}

						//now, look to see if the point has left the hysteresis band out
						//the bottom. if so, then we have enough information to calculate
						//a zero-cross point.
						if(*(pS - 1) >= dLo && *pS < dLo && bInCut == true)
						{
							//instruct the linear regression section below that it's time
							//to find a midpoint, set the zero cross finder to look for a 
							//low to high transition now, and turn off the "inside" flag
							bCalc = true;
							bTrans = true;
							bInCut = false;
						}
					} 
					
					//if we're not looking for high-to-low, we must be looking for low-to-high.
					//this code is pretty much a carbon-copy duplicate of the above with some
					//signs reversed and variables changed. note that since this is an else
					//statement, if the above code executed, this code will not
					else 
					{
						//was there a transition from low to inside? then set the "inside" 
						//flag to true and hold on. or, was there a transition from inside to
						//low? in that case, we've left the hysteresis band without actually 
						//getting a zero cross point, so reset the "inside" flag.
						if(*(pS - 1) <= dLo && *pS > dLo)
						{
							//use the point just outside the transition as the first point 
							//used for calculation. save both a pointer to the current item 
							//and a counter reflecting the current item
							pJ = pS - 1;
							J = I - 1;
							bInCut = true;
						}
						else if(*(pS - 1) > dLo && *pS <= dLo && bInCut == true)
						{
							bInCut = false;
						}

						//now, look to see if the point has left the hysteresis band out
						//the top. if so, then we have enough information to calculate
						//a zero-cross point.
						if(*(pS - 1) <= dHi && *pS > dHi && bInCut == true)
						{
							//instruct the linear regression section below that it's time
							//to find a midpoint, set the zero cross finder to look for a 
							//high to low transition now, and turn off the "inside" flag
							bCalc = true;
							bTrans = false;
							bInCut = false;
						}
					}

					//has either one of the above code sections set aside a enough
					//points to find a zero-cross point?
					if(bCalc == true)
					{
						//we'll use the current point (the point that just left
						//the hysteresis band) as our final point. this way, even
						//if there are zero points in the hysteresis band, we'll
						//still have a total of 2 points to calculate a regression.
						
						//build an array of all points from the beginning (J) to the
						//end (currently I)
						oL = new nfMath.LinkedList();
						for(K = 0; K <= (I - J); K++)
							oL.Push(*pJ++);

						//the linked list oL now contains all points between (and
						//including) I and J. convert it to a standard array and
						//use the LinearRegression class to return a linear model
						//between the two points and including all points between
						dList = new double[oL.Count];
						oL.CopyTo(dList, 0);

						//initialize the linear regression model, assign its
						//values, and return the correctly built regression model
						//starting with I as the initial Y point
						oReg = new nfMath.PolyRegression(1);
						oReg.setSource(dList, J);
						mParam = oReg.Model();
						
						//finally, assign this point using what we found with the
						//linear regression model and push this result onto the
						//linked list containing all midpoints
						pCur.X = (yMean - mParam[1, 0]) / mParam[0, 0];
						pCur.Y = yMean;
						oMid.Push(pCur);

						//don't need to find a point anymore
						bCalc = false;
					}
 				}
			}

			//at least 5 midpoints are required in order to successfully derive a single
			//tooth (since one midpoint is required to lead, one is required to end, and
			//three are required to derive a single tooth). if we don't successfully find
			//a single tooth, return 0 as the last point (so the next iteration can use
			//all the points in this one)
			if(oMid.Count < 5)
			{
				return null;
			}

			//since we're done adding new midpoints to the linked list, it's now
			//advantageous to convert the list into a true array. then, since there
			//will be exactly one less extrema (min or max) than midpoints, we can
			//size that array too
			dMid = new Point[oMid.Count];
			dExt = new Point[oMid.Count - 1];
			oMid.CopyTo(dMid, 0);

			//loop through all midpoints to create a list of extremas
			for(I = 0; I < dExt.Length; I++)
			{
				//find the width (in x units) of the current half-cycle
				dDelta = dMid[I + 1].X - dMid[I].X;
				pCur = new nfABS.Point();

				//what type of calculations to find the high point for this extrema?
				//remember: xCut specifies how much of the signal on either side of
				//the midpoint will be ignored. for example, for a .8 xCut:
				//
				// ---|----**----|----**----|----**----|----**----|----**----|---
				//
				//only the astersisk'd areas will be used for calculation of extrema
				switch(CuspType)
				{
					case eCuspType.Average:
						//for an average calculation, we want to use only the points within 
						//the x cut (not the points just barely on the outside) the theory is,
						//with the average, if two points are below the midrange line and one 
						//is above, there's a possibility of establishing an extrema on the 
						//wrong side of the midline
						iLo = (int)Math.Ceiling(dMid[I].X + 0.5 * dDelta * xCut);
						iHi = (int)Math.Floor(dMid[I + 1].X - 0.5 * dDelta * xCut);
						dList = new double[iHi - iLo + 1];

						//now, add the points between iLo and iHi (inclusive) to a list for
						//averaging together.
						for(J = iLo; J <= iHi; J++)
						{ dList[J - iLo] = dData[J]; }
						
						//set the average at the exact horizontal center between two mids
						//and at the correct vertical average
						pCur.X = (dMid[I + 1].X + dMid[I].X) / 2;
						pCur.Y = nfMath.Basic.Average(dList);
						break;

					case eCuspType.Quadratic:
						//for a quadratic calculation, it's ok to use the points that
						//are actually below the midpoints. IMPORTANT NOTE: it is HIGHLY
						//recommended that with a Quadratic cusp model, the xCut be set
						//to zero. This will guarantee that each peak receives three or
						//more points and will prevent any software exceptions. Using an
						//xCut can provide more accurate data, but the signal better not be
						//noisy, or else math exceptions can occur.
						iLo = (int)Math.Floor(dMid[I].X + 0.5 * dDelta * xCut);
						iHi = (int)Math.Ceiling(dMid[I + 1].X - 0.5 * dDelta * xCut);
						
						//create the array and populate it
						dList = new double[iHi - iLo + 1];

						//now, add the points between iLo and iHi (inclusive) to a list for
						//quadratic regression.
						for(J = iLo; J <= iHi; J++)
							{ dList[J - iLo] = dData[J]; }
						
						//build a parabola matching this data.
						oReg = new nfMath.PolyRegression(2);
						oReg.setSource(dList, iLo);
						mParam = oReg.Model();
						
						//finally, find the maximum of this parabola by using some simple
						//calculus.
						//
						//      Y = AX^2 + BX + C			dY = 2AX + B
						//      0 = 2AX + B					-B = 2AX
						//						-B/2A = X
						//
						//if the first paramater (A) is zero, then we need to cheat and sneak
						//around it by using the same calculation as average for X
						if(mParam[0, 0] == 0)
						{	
							pCur.X = (dMid[I + 1].X + dMid[I].X) / 2;
						} else {
							pCur.X = -mParam[1, 0] / (2 * mParam[0, 0]); 
						}
						pCur.Y = mParam[0, 0] * pCur.X * pCur.X + mParam[1, 0] * pCur.X + mParam[2, 0];
						break;
				}

				//finally, assign the current extreme to the value specified.
				dExt[I] = pCur;
			}










			//finally, create our list of TTL high and low transfer points.
			dTTL = new Point[(dMid.Length - 2) * 2];
			bTrans = true;

			//loop through all midpoints to create a list of TTL transfers
			for(I = 0; I < dMid.Length - 2; I++)
			{
				//first, figure out the current cycle's amplitude. the 10% or 90% factor
				//relates only to the current cycle transition, not globally
				dDelta = dExt[I + 1].Y - dExt[I].Y;

				//next, define the two points that determine when we're above and below
				//our TTL limits.
				dLo = dExt[I].Y + dDelta * ttlLo;
				dHi = dExt[I].Y + dDelta * ttlHi;









				 
				//are we currently looking for a high-to-low transition?
				if(bTrans == false)
				{
					//was there a transition from high to inside? then set the "inside" 
					//flag to true and hold on. or, was there a transition from inside to
					//high? in that case, we've left the hysteresis band without actually 
					//getting a zero cross point, so reset the "inside" flag.
					if(*(pS - 1) >= dHi && *pS < dHi)
					{
						//use the point just outside the transition as the first point 
						//used for calculation. save both a pointer to the current item 
						//and a counter reflecting the current item
						pJ = pS - 1;
						J = I - 1;
						bInCut = true;
					}
					else if(*(pS - 1) < dHi && *pS >= dHi && bInCut == true)
					{
						bInCut = false;
					}

					//now, look to see if the point has left the hysteresis band out
					//the bottom. if so, then we have enough information to calculate
					//a zero-cross point.
					if(*(pS - 1) >= dLo && *pS < dLo && bInCut == true)
					{
						//instruct the linear regression section below that it's time
						//to find a midpoint, set the zero cross finder to look for a 
						//low to high transition now, and turn off the "inside" flag
						bCalc = true;
						bTrans = true;
						bInCut = false;
					}
				} 
					
					//if we're not looking for high-to-low, we must be looking for low-to-high.
					//this code is pretty much a carbon-copy duplicate of the above with some
					//signs reversed and variables changed. note that since this is an else
					//statement, if the above code executed, this code will not
				else 
				{
					//was there a transition from low to inside? then set the "inside" 
					//flag to true and hold on. or, was there a transition from inside to
					//low? in that case, we've left the hysteresis band without actually 
					//getting a zero cross point, so reset the "inside" flag.
					if(*(pS - 1) <= dLo && *pS > dLo)
					{
						//use the point just outside the transition as the first point 
						//used for calculation. save both a pointer to the current item 
						//and a counter reflecting the current item
						pJ = pS - 1;
						J = I - 1;
						bInCut = true;
					}
					else if(*(pS - 1) > dLo && *pS <= dLo && bInCut == true)
					{
						bInCut = false;
					}

					//now, look to see if the point has left the hysteresis band out
					//the top. if so, then we have enough information to calculate
					//a zero-cross point.
					if(*(pS - 1) <= dHi && *pS > dHi && bInCut == true)
					{
						//instruct the linear regression section below that it's time
						//to find a midpoint, set the zero cross finder to look for a 
						//high to low transition now, and turn off the "inside" flag
						bCalc = true;
						bTrans = false;
						bInCut = false;
					}
				}

				//has either one of the above code sections set aside a enough
				//points to find a zero-cross point?
				if(bCalc == true)
				{
					//we'll use the current point (the point that just left
					//the hysteresis band) as our final point. this way, even
					//if there are zero points in the hysteresis band, we'll
					//still have a total of 2 points to calculate a regression.
						
					//build an array of all points from the beginning (J) to the
					//end (currently I)
					oL = new nfMath.LinkedList();
					for(K = 0; K <= (I - J); K++)
						oL.Push(*pJ++);

					//the linked list oL now contains all points between (and
					//including) I and J. convert it to a standard array and
					//use the LinearRegression class to return a linear model
					//between the two points and including all points between
					dList = new double[oL.Count];
					oL.CopyTo(dList, 0);

					//initialize the linear regression model, assign its
					//values, and return the correctly built regression model
					//starting with I as the initial Y point
					oReg = new nfMath.PolyRegression(1);
					oReg.setSource(dList, J);
					mParam = oReg.Model();
						
					//finally, assign this point using what we found with the
					//linear regression model and push this result onto the
					//linked list containing all midpoints
					pCur.X = (yMean - mParam[1, 0]) / mParam[0, 0];
					pCur.Y = yMean;
					oMid.Push(pCur);

					//don't need to find a point anymore
					bCalc = false;
				}
			}

			return dExt;
		}
	}
}